@@ -371,7 +371,7 @@ internal class RecordType
 
 namespace Beagle.Filters {
     
-	public class FilterPPT : FilterOle {
+	public class FilterPPT : Beagle.Daemon.Filter {
 
 		private enum TextType {
 			Invalid = -1,
@@ -386,14 +386,19 @@ namespace Beagle.Filters {
 			QuarterBody
 		};
 
+		Infile file;
 		TextType textType;
+		string FileName;
 		public FilterPPT () 
 		{
 			AddSupportedMimeType ("application/vnd.ms-powerpoint");
 			textType = TextType.Invalid;
+			file = null;
+			FileName = null;
+			SnippetMode = true;
 		}
 
-		private int ParseElement (Gsf.Input stream)
+		private int ParseElement (Input stream)
 		{
 			byte [] data = stream.Read(8);
 			if (data == null)
@@ -452,7 +457,7 @@ namespace Beagle.Filters {
 						if (IsHot)
 							HotDown ();
 						else
-							AppendWhiteSpace ();
+							AppendStructuralBreak ();
 					}  else if (opcode == RecordType.TypeCode.TextHeaderAtom) {
 						data = stream.Read (4);
 						textType = (TextType) GetInt32 (data, 0);
@@ -475,13 +480,13 @@ namespace Beagle.Filters {
 			if (sumStream != null)
 				sumMeta = Msole.MetadataReadReal (sumStream);
 			else
-				Logger.Log.Error ("SummaryInformationStream not found");
-			
+				Logger.Log.Error ("SummaryInformationStream not found in {0}", FileName);
+
 			DocMetaData docSumMeta = null;
 			if (docSumStream != null)
 				docSumMeta = Msole.MetadataReadReal (docSumStream);
 			else
-				Logger.Log.Error ("DocumentSummaryInformationStream not found");
+				Logger.Log.Error ("DocumentSummaryInformationStream not found in {0}", FileName);
 
 			DocProp prop = null;
 			string str = null;
@@ -545,37 +550,120 @@ namespace Beagle.Filters {
 			int childCount = 0;
 			int found = 0;
 			
-			childCount = file.NumChildren();
-			for (int i = 0; i < childCount && found != 2; i++) {
-				str = file.NameByIndex (i);
-				if (string.Compare (str, "SummaryInformation") == 0) {
-					sumStream = file.ChildByIndex (i);
-					found = 1;
+			if (file == null) {
+				Finished ();
+				return;
+			}
+			
+			try {
+				// FIXME: Should try to use Encoding instead of 
+				// string.IndexOf ()... Hacky stuff ;-)
+				childCount = file.NumChildren();
+				for (int i = 0; i < childCount && found != 2; i++) {
+					str = file.NameByIndex (i);	
+					if (str.IndexOf ("SummaryInformation") > -1 && found < 1) {
+						sumStream = file.ChildByIndex (i);
+						found = 1;
+					}
+					else if (str.IndexOf ("DocumentSummaryInformation") > -1) {
+						docSumStream = file.ChildByIndex (i);
+						found = 2;
+					}
 				}
-				else if (string.Compare (str, "DocumentSummaryInformation") == 0) {
-					docSumStream = file.ChildByIndex (i);
-					found = 2;
-				}
+				ExtractMetaData (sumStream, docSumStream);
+			} catch (Exception e) {
+				Logger.Log.Error ("Exception {0} occurred duing DoPullProperties.", e.Message);
+				Finished ();
 			}
-			ExtractMetaData (sumStream, docSumStream);
 		}
 
 		override protected void DoPull ()
 		{
-			Input stream = file.ChildByName ("PowerPoint Document");
+			if (file == null) {
+				Finished ();
+				return;
+			}
 
-			// The parsing was getting terminated when "EndDocument"
-			// container was parsed.  We need to continue our 
-			// parsing till the end of the file, since, some of the
-			// slides do persist after the actual "Document" 
-			// container.
-			// PPTs exported from OO.o actually writes almost all the slides
-			// after "Document" container.
-			// And certain PPTs do have some slides in after
-			// "Document" container.
-			while (!stream.Eof)
-				ParseElement(stream);
+			Input stream = null;
+			try {
+				stream = file.ChildByName ("PowerPoint Document");
+
+				if (stream != null) {
+
+					// The parsing was getting terminated when "EndDocument"
+					// container was parsed.  We need to continue our 
+					// parsing till the end of the file, since, some of the
+					// slides do persist after the actual "Document" 
+					// container.
+					// PPTs exported from OO.o actually writes almost all the slides
+					// after "Document" container.
+					// And certain PPTs do have some slides in after
+					// "Document" container.
+					while (!stream.Eof)
+						ParseElement (stream);
+				} else {
+					Logger.Log.Error ("Ole stream not found in {0}.  Content extraction skipped.", FileName);
+				}
 			Finished();
+			} catch (Exception e) {
+				Logger.Log.Error ("Exception {0} occurred during DoPull.", e.Message);
+				Finished ();
+			}
 		}
+
+		override protected void DoOpen (FileInfo info)
+		{
+			FileName = info.FullName;
+
+			try {
+				Gsf.Global.Init ();
+				Input input = Input.MmapNew (info.FullName);
+				if (input != null) {
+					input = input.Uncompress();
+					file = new InfileMSOle (input);
+				}
+				if (input == null || file == null) {
+					Logger.Log.Error ("Unable to open [{0}] ",info.FullName);
+					Finished ();
+					return;
+				}
+				
+			} catch (Exception e) {
+				Logger.Log.Error ("Unable to open "+info.FullName);
+				Finished ();
+				return;
+			}
+			
+			// PPT 95/97-2000 format contains a "PP97_DUALSTORAGE", which is required
+			// to index PPT 97-2000 files.
+			// We don't support PPT 95 files, however, we happily accept patches ;-)
+
+			Input dualStorTemp = null;
+			try {
+				if ((dualStorTemp = file.ChildByName ("PP97_DUALSTORAGE")) != null) {
+					// "PP97_DUALSTORAGE" is a storage containing some streams
+					if (dualStorTemp.Handle != IntPtr.Zero)
+						file = (Gsf.Infile) GLib.Object.GetObject (dualStorTemp.Handle);
+				} else if (((dualStorTemp = file.ChildByName ("Header")) != null) ||
+					   ((dualStorTemp = file.ChildByName ("PowerPoint Document")) == null)) {
+					Logger.Log.Error ("{0} is a PPT 95/4.0 file.  Beagle does not support PPT 95 files. Skipping...", FileName);
+					Finished ();
+				}
+			} catch (Exception e) {
+				
+				Logger.Log.Error ("Unable to open OleFile stream of "+info.FullName);
+				Finished ();
+			}
+		}
+
+		// FIXME: These are utility functions and can be useful 
+		// outside this filter as well.
+		public static int GetInt32 (byte [] data, int offset) {
+			return data[offset] + (data[offset + 1] << 8) + (data[offset + 2] << 16) + (data[offset + 3] << 24);
+		}
+		public static int GetInt16 (byte [] data, int offset) {
+			return data[offset] + (data[offset + 1] << 8);
+		}
+
 	}
 }
