@@ -16,37 +16,53 @@ using System.Text;
 namespace Beagle.Util {
 namespace Camel {
 	
-	public enum CamelFlags {
-		ANSWERED     = 1<<0,
-		DELETED      = 1<<1,
-		DRAFT        = 1<<2,
-		FLAGGED      = 1<<3,
-		SEEN         = 1<<4,
-		ATTACHMENTS  = 1<<5,
-		ANSWERED_ALL = 1<<6,
-		UNKNOWN_7    = 1<<7,
-		UNKNOWN_8    = 1<<8
+	public enum CamelFlags : uint {
+		Answered     = 1 << 0,
+		Deleted      = 1 << 1,
+		Draft        = 1 << 2,
+		Flagged      = 1 << 3,
+		Seen         = 1 << 4,
+		Attachments  = 1 << 5,
+		AnsweredAll  = 1 << 6,
+		Junk         = 1 << 7,
+		Secure       = 1 << 8
 	}
 	
 public abstract class Summary : IEnumerable { 
 	public SummaryHeader header;
 	internal string filename;
 
-	public static Summary load (string file) 
+	public static Summary LoadMBoxSummary (string file)
 	{
-		Summary s;
-		if (file.EndsWith ("/summary")) { 
-			s = new ImapSummary ();
-		} else {
-			s = new MBoxSummary ();
-		}
-		s.filename = file;
-		FileStream f = File.OpenRead (file);
-		s.header = s.ReadHeader (f);
-		f.Close ();
-		
+		Summary s = new MBoxSummary ();
+		s.Load (file);
+
 		return s;
 	}
+
+	public static Summary LoadImapSummary (string file)
+	{
+		Summary s = new ImapSummary ();
+		s.Load (file);
+
+		return s;
+	}
+
+	public static Summary LoadImap4Summary (string file)
+	{
+		Summary s = new Imap4Summary ();
+		s.Load (file);
+
+		return s;
+	}
+
+	private void Load (string file)
+	{
+		this.filename = file;
+		FileStream f = File.OpenRead (file);
+		this.header = this.ReadHeader (f);
+		f.Close ();
+	}
 	
 	public IEnumerator GetEnumerator () {
 		return new SummaryEnumerator (this);
@@ -155,10 +171,26 @@ public abstract class Summary : IEnumerable {
 
 		protected override MessageInfo ReadMessageInfo (FileStream f)
 		{
-			return new ImapMessageInfo (f);
+			return new ImapMessageInfo (f, true);
 		}
  	}
 
+	public class Imap4Summary : Summary {
+		public Imap4Summary ()
+		{
+		}
+
+		protected override SummaryHeader ReadHeader (FileStream f)
+		{
+			return new ImapSummaryHeader (f);
+		}
+
+		protected override MessageInfo ReadMessageInfo (FileStream f)
+		{
+			return new ImapMessageInfo (f, false);
+		}
+	}
+
 	public class MessageInfo {
 		public string uid, subject, from, to, cc, mlist;
 		public uint size, flags;
@@ -167,16 +199,16 @@ public abstract class Summary : IEnumerable {
 
 		private void SkipContentInfo (FileStream f)
 		{
-			Decode.Token (f); // type
-			Decode.Token (f); // subtype
+			Decode.SkipToken (f); // type
+			Decode.SkipToken (f); // subtype
 			uint count = Decode.UInt (f); // count
 			for (int i = 0; i < count; ++i) {
-				Decode.Token (f); // name
-				Decode.Token (f); // value
+				Decode.SkipToken (f); // name
+				Decode.SkipToken (f); // value
 			}
-			Decode.Token (f); // id
-			Decode.Token (f); // description
-			Decode.Token (f); // encoding
+			Decode.SkipToken (f); // id
+			Decode.SkipToken (f); // description
+			Decode.SkipToken (f); // encoding
 			Decode.UInt (f); // size
 
 			count = Decode.UInt (f); // child count
@@ -197,8 +229,8 @@ public abstract class Summary : IEnumerable {
 			cc       = Decode.String (f);
 			mlist    = Decode.String (f);
 
-			Decode.FixedInt (f);
-			Decode.FixedInt (f);
+			Decode.SkipFixedInt (f);
+			Decode.SkipFixedInt (f);
 
 			uint count;
 
@@ -206,8 +238,8 @@ public abstract class Summary : IEnumerable {
 			count = Decode.UInt (f);
 			if (count > 0) {
 				for (int i = 0; i < count; i++) {
-					Decode.FixedInt (f);
-					Decode.FixedInt (f);
+					Decode.SkipFixedInt (f);
+					Decode.SkipFixedInt (f);
 				}
 			}
 
@@ -215,7 +247,7 @@ public abstract class Summary : IEnumerable {
 			count = Decode.UInt (f);
 			if (count > 0) {
 				for (int i = 0; i < count; i++) {
-					Decode.String (f);
+					Decode.SkipString (f);
 				}
 			}
 
@@ -223,8 +255,8 @@ public abstract class Summary : IEnumerable {
 			count = Decode.UInt (f);
 			if (count > 0){
 				for (int i = 0; i < count; i++){
-					Decode.String (f);
-					Decode.String (f);
+					Decode.SkipString (f);
+					Decode.SkipString (f);
 				}
 			}
 
@@ -243,35 +275,43 @@ public abstract class Summary : IEnumerable {
 		
 		private bool CheckFlag (CamelFlags test)
 		{
-			return (flags & (int) test) == (int) test;
+			return (flags & (uint) test) == (uint) test;
 		}
 		
 		public bool IsAnswered {
-			get { return CheckFlag (CamelFlags.ANSWERED); }
+			get { return CheckFlag (CamelFlags.Answered); }
 		}
 
 		public bool IsDeleted {
-			get { return CheckFlag (CamelFlags.DELETED); }
+			get { return CheckFlag (CamelFlags.Deleted); }
 		}
 
 		public bool IsDraft {
-			get { return CheckFlag (CamelFlags.DRAFT); }
+			get { return CheckFlag (CamelFlags.Draft); }
 		}
 
 		public bool IsFlagged {
-			get { return CheckFlag (CamelFlags.FLAGGED); }
+			get { return CheckFlag (CamelFlags.Flagged); }
 		}
 
 		public bool IsSeen {
-			get { return CheckFlag (CamelFlags.SEEN); }
+			get { return CheckFlag (CamelFlags.Seen); }
 		}
 
 		public bool HasAttachments {
-			get { return CheckFlag (CamelFlags.ATTACHMENTS); }
+			get { return CheckFlag (CamelFlags.Attachments); }
 		}
 
 		public bool IsAnsweredAll {
-			get { return CheckFlag (CamelFlags.ANSWERED_ALL); }
+			get { return CheckFlag (CamelFlags.AnsweredAll); }
+		}
+
+		public bool IsJunk {
+			get { return CheckFlag (CamelFlags.Junk); }
+		}
+
+		public bool IsSecure {
+			get { return CheckFlag (CamelFlags.Secure); }
 		}
 	}
 
@@ -295,10 +335,12 @@ public abstract class Summary : IEnumerable {
 	public class ImapMessageInfo : MessageInfo {
 		public uint server_flags;
 		
-		public ImapMessageInfo (FileStream f) : base (f)
+		public ImapMessageInfo (FileStream f, bool content_info_load) : base (f)
 		{
 			server_flags = Decode.UInt (f);
-			PerformContentInfoLoad (f);
+
+			if (content_info_load)
+				PerformContentInfoLoad (f);
 		}
 
 		public override string ToString ()
@@ -328,15 +370,13 @@ public abstract class Summary : IEnumerable {
 
 		private bool ContentInfoLoad (FileStream f)
 		{
-		    string token;
-
 		    if (f.ReadByte () == 0) 
 			return true;
 
 		    // type
-		    token = Decode.Token (f);
+		    Decode.SkipToken (f);
 		    // subtype
-		    token = Decode.Token (f);
+		    Decode.SkipToken (f);
 
 		    uint count;
 		    count = Decode.UInt (f);
@@ -344,19 +384,19 @@ public abstract class Summary : IEnumerable {
 			return false;
 		    for (int i = 0; i < count; i++) {
 			// Name
-			token = Decode.Token (f);
+			Decode.SkipToken (f);
 			// Value
-			token = Decode.Token (f);
+			Decode.SkipToken (f);
 		    }
 		    
 		    // id
-		    token = Decode.Token (f);
+		    Decode.SkipToken (f);
 
 		    // description
-		    token = Decode.Token (f);
+		    Decode.SkipToken (f);
 
 		    // encoding
-		    token = Decode.Token (f);
+		    Decode.SkipToken (f);
 
 		    // size
 		    Decode.UInt (f);
@@ -376,16 +416,30 @@ public abstract class Summary : IEnumerable {
 		
 		public SummaryHeader (FileStream f)
 		{
+			bool legacy;
+
 			version = Decode.FixedInt (f);
+
+			if (version > 0xff && (version & 0xff) < 12)
+				throw new Exception ("Summary header version too low");
+
+			if (version < 0x100 && version >= 13)
+				legacy = false;
+			else
+				legacy = true;
+
 			flags   = Decode.FixedInt (f);
 			nextuid = Decode.FixedInt (f);
 			time    = Decode.Time (f);
 			count   = Decode.FixedInt (f);
-			unread  = Decode.FixedInt (f);
-			deleted = Decode.FixedInt (f);
-			junk    = Decode.FixedInt (f);
 
-			//Console.WriteLine ("V={0} time={1}, count={2} unread={3} deleted={4} junk={5}", version, time, count, unread, deleted, junk);
+			if (!legacy) {
+				unread  = Decode.FixedInt (f);
+				deleted = Decode.FixedInt (f);
+				junk    = Decode.FixedInt (f);
+			}
+
+			//Console.WriteLine ("V={0} ({1}) time={2}, count={3} unread={4} deleted={5} junk={6}", version, version & 0xff, time, count, unread, deleted, junk);
 		}
 	}
 
@@ -399,6 +453,8 @@ public abstract class Summary : IEnumerable {
 			local_version = Decode.FixedInt (f);
 			mbox_version = Decode.FixedInt (f);
 			folder_size = Decode.FixedInt (f);
+
+			//Console.WriteLine ("local_version={0}  mbox_version={1}  folder_size={2}", local_version, mbox_version, folder_size);
 		}
 	}
 
@@ -406,14 +462,28 @@ public abstract class Summary : IEnumerable {
 		
 		public ImapSummaryHeader (FileStream f) : base (f)
 		{
-			int version = Decode.FixedInt (f);
+			// Check for legacy version
+			if (base.version != 0x30c) { // 780
+				int version = Decode.FixedInt (f);
 
-			// Right now we only support summary versions 1 and 3
-			// (which according to fejj are identical)
-			if (version != 1 && version != 3)
-				throw new Exception ("Incompatible summary version: " + version);
+				//Console.WriteLine ("imap version={0}", version);
 
-			int validity = Decode.FixedInt (f);
+				if (version < 0)
+					throw new Exception ("IMAP summary version too low");
+
+				// Right now we only support summary versions 1 through 3
+				if (version > 3)
+					throw new Exception (String.Format ("Reported summary version ({0}) is too new", version));
+
+				if (version == 2)
+					Decode.FixedInt (f);
+
+				// validity
+				Decode.SkipFixedInt (f);
+			} else {
+				// validity
+				Decode.UInt (f);
+			}
 		}
 	}
 	
@@ -444,6 +514,14 @@ public abstract class Summary : IEnumerable {
 			return new System.String (e.GetChars (buffer, 0, len));
 		    }
 		}
+
+		public static void SkipToken (FileStream f)
+		{
+			int len = (int) UInt (f);
+			len -= 32;
+			if (len > 0)
+				f.Seek (len, SeekOrigin.Current);
+		}
 	       
 		public static string String (FileStream f)
 		{
@@ -457,6 +535,14 @@ public abstract class Summary : IEnumerable {
 			return new System.String (e.GetChars (buffer, 0, len));
 		}
 
+		public static void SkipString (FileStream f)
+		{
+			int len = (int) UInt (f);
+			--len;
+			if (len > 0)
+				f.Seek (len, SeekOrigin.Current);
+		}
+
 		public static uint UInt (FileStream f)
 		{
 			uint value = 0;
@@ -478,6 +564,11 @@ public abstract class Summary : IEnumerable {
 			return (b [0] << 24) | (b [1] << 16) | (b [2] << 8) | b [3];
 		}
 
+		public static void SkipFixedInt (FileStream f)
+		{
+			f.Seek (4, SeekOrigin.Current);
+		}
+
 		public static DateTime Time (FileStream f)
 		{
 			byte [] b = new byte [4];
@@ -501,6 +592,7 @@ public abstract class Summary : IEnumerable {
 		}
 	}
 
+#if false
 	class Test {
 		void Main (string [] args)
 		{
@@ -511,7 +603,7 @@ public abstract class Summary : IEnumerable {
 			else
 				file = args [0];
 			
-			Summary s = Summary.load (file);
+			Summary s = Summary.LoadMBoxSummary (file);
 			foreach (MessageInfo m in s) {
 				Console.WriteLine(m);
 			}
@@ -519,5 +611,6 @@ public abstract class Summary : IEnumerable {
 		}
 		
 	}
+#endif
 }
 }
