@@ -36,7 +36,6 @@ using Beagle.Util;
 using Beagle.Daemon;
 
 using Camel = Beagle.Util.Camel;
-using GConf;
 using Mono.Posix;
 
 namespace Beagle.Daemon.EvolutionMailDriver {
@@ -251,6 +250,49 @@ namespace Beagle.Daemon.EvolutionMailDriver {
 		{
 			return this.CrawlFile.FullName.GetHashCode ();
 		}
+
+		// FIXME FIXME FIXME!
+		// This is a fragile piece-of-shit parser.
+		// This will not do the right thing if the address is quoted
+		// or contains special characters.  For example, we will do the wrong thing
+		// with:
+		// "Foo, bar" <foo@bar.com>, Zoo <zoo@bar.com>"
+		// Also, this almost certainly won't deal well w/ broken addresses.
+		static protected ICollection ExtractAddresses (string addr_str)
+		{
+			string [] split = addr_str.Split (',');
+
+			ArrayList addresses = new ArrayList ();
+			foreach (string part in split) {
+				if (part == "")
+					continue;
+				if (part [part.Length - 1] == '>') {
+					int i = part.LastIndexOf ('<');
+					addresses.Add (part.Substring (i+1, part.Length-i-2));
+				} else if (part [part.Length - 1] == ')') {
+					int i = part.LastIndexOf ('(');
+					addresses.Add (part.Substring (i+1, part.Length-i-2));
+				} else {
+					addresses.Add (part);
+				}
+			}
+			
+			return addresses;
+		}
+
+		static protected void MapAddressStringToProperties (Indexable indexable,
+								    string    property_key,
+								    string    address_string)
+		{
+			if (address_string == null || address_string.Length == 0)
+				return;
+
+			foreach (string one_addr in ExtractAddresses (address_string)) {
+				Property prop;
+				prop = Property.NewKeyword (property_key, one_addr);
+				indexable.AddProperty (prop);
+			}
+		}
 	}
 
 	public class EvolutionMailIndexableGeneratorMbox : EvolutionMailIndexableGenerator {
@@ -381,39 +423,70 @@ namespace Beagle.Daemon.EvolutionMailDriver {
 					return null;
 				}
 				
-				string uid_str = x_evolution.Substring (0, x_evolution.IndexOf ('-'));
+				int separator_idx = x_evolution.IndexOf ('-');
+
+				string uid_str = x_evolution.Substring (0, separator_idx);
 				string uid = Convert.ToUInt32 (uid_str, 16).ToString (); // ugh.
+				uint flags = Convert.ToUInt32 (x_evolution.Substring (separator_idx + 1), 16);
 				
-				Indexable indexable = this.GMimeMessageToIndexable (uid, message);
+				Indexable indexable = this.GMimeMessageToIndexable (uid, message, flags);
+				
+				if (indexable == null)
+					return null;
+
 				++this.indexed_count;
+
+				// HACK: update your recipients
+				EvolutionMailQueryable.AddAsYourRecipient (indexable);
 				
 				return indexable;
 			}
 		}
 
-		private Indexable GMimeMessageToIndexable (string uid, GMime.Message message)
+		private static bool CheckFlags (uint flags, Camel.CamelFlags test)
 		{
+			return (flags & (uint) test) == (uint) test;
+		}
+
+		private Indexable GMimeMessageToIndexable (string uid, GMime.Message message, uint flags)
+		{
+			// Don't index messages flagged as junk
+			if (CheckFlags (flags, Camel.CamelFlags.Junk))
+				return null;
+
 			System.Uri uri = EvolutionMailQueryable.EmailUri (this.account_name, this.folder_name, uid);
 			Indexable indexable = new Indexable (uri);
 
 			indexable.Timestamp = message.Date;
 			indexable.Type = "MailMessage";
-			indexable.MimeType = "text/plain"; // FIXME
+			indexable.MimeType = "text/plain";
+			indexable.CacheContent = false;
 
-			indexable.AddProperty (Property.New ("fixme:client", "evolution"));
+			indexable.AddProperty (Property.NewKeyword ("fixme:client", "evolution"));
 
-			indexable.AddProperty (Property.NewKeyword ("dc:title",       GMime.Utils.HeaderDecodePhrase (message.Subject)));
+			string subject = GMime.Utils.HeaderDecodePhrase (message.Subject);
+			indexable.AddProperty (Property.New ("dc:title", subject));
+			indexable.AddProperty (Property.New ("fixme:subject", subject));
 
-			indexable.AddProperty (Property.NewKeyword ("fixme:account",  "Local"));
-                        indexable.AddProperty (Property.NewKeyword ("fixme:folder",   this.folder_name));
-			indexable.AddProperty (Property.NewKeyword ("fixme:subject",  GMime.Utils.HeaderDecodePhrase (message.Subject)));
-                        indexable.AddProperty (Property.NewKeyword ("fixme:to",       message.GetRecipientsAsString (GMime.Message.RecipientType.To)));
-			indexable.AddProperty (Property.NewKeyword ("fixme:from",     GMime.Utils.HeaderDecodePhrase (message.Sender)));
-			indexable.AddProperty (Property.NewKeyword ("fixme:cc",       message.GetRecipientsAsString (GMime.Message.RecipientType.Cc)));
+			indexable.AddProperty (Property.NewKeyword ("fixme:account", "Local"));
+                        indexable.AddProperty (Property.NewKeyword ("fixme:folder", this.folder_name));
 
-			if (this.folder_name == "Sent")
+			string to_str = message.GetRecipientsAsString (GMime.Message.RecipientType.To);
+			string cc_str = message.GetRecipientsAsString (GMime.Message.RecipientType.Cc);
+			string from_str = GMime.Utils.HeaderDecodePhrase (message.Sender);
+                        indexable.AddProperty (Property.NewKeyword ("fixme:to", to_str));       
+			indexable.AddProperty (Property.NewKeyword ("fixme:cc", cc_str));
+			indexable.AddProperty (Property.NewKeyword ("fixme:from", from_str));
+
+			if (this.folder_name == "Sent") {
 				indexable.AddProperty (Property.NewFlag ("fixme:isSent"));
 
+				MapAddressStringToProperties (indexable, "fixme:sentTo", to_str);
+				MapAddressStringToProperties (indexable, "fixme:sentTo", cc_str);
+			} else {
+				MapAddressStringToProperties (indexable, "fixme:gotFrom", from_str);
+			}
+
 			if (message.MimePart is GMime.Multipart || message.MimePart is GMime.MessagePart)
 				indexable.AddProperty (Property.NewFlag ("fixme:hasAttachments"));
 
@@ -428,36 +501,27 @@ namespace Beagle.Daemon.EvolutionMailDriver {
 				indexable.AddProperty (Property.NewDate ("fixme:sentdate", message.Date));
 			else
 				indexable.AddProperty (Property.NewDate ("fixme:received", message.Date));
-			
-#if false
-			// FIXME - XXX
-                        indexable.AddProperty (Property.NewKeyword ("fixme:flags",    messageInfo.flags));
 
-			if (messageInfo.received != DateTime.MinValue)
-				indexable.AddProperty (Property.NewDate ("fixme:received", messageInfo.received));
+			indexable.AddProperty (Property.NewKeyword ("fixme:flags", flags));
 
-			if (messageInfo.sent != DateTime.MinValue)
-				indexable.AddProperty (Property.NewDate ("fixme:sentdate", messageInfo.sent));
-
-			if (messageInfo.IsAnswered)
+			if (CheckFlags (flags, Camel.CamelFlags.Answered))
 				indexable.AddProperty (Property.NewFlag ("fixme:isAnswered"));
 
-			if (messageInfo.IsDeleted)
+			if (CheckFlags (flags, Camel.CamelFlags.Deleted))
 				indexable.AddProperty (Property.NewFlag ("fixme:isDeleted"));
 
-			if (messageInfo.IsDraft)
+			if (CheckFlags (flags, Camel.CamelFlags.Draft))
 				indexable.AddProperty (Property.NewFlag ("fixme:isDraft"));
 
-			if (messageInfo.IsFlagged)
+			if (CheckFlags (flags, Camel.CamelFlags.Flagged))
 				indexable.AddProperty (Property.NewFlag ("fixme:isFlagged"));
 
-			if (messageInfo.IsSeen)
+			if (CheckFlags (flags, Camel.CamelFlags.Seen))
 				indexable.AddProperty (Property.NewFlag ("fixme:isSeen"));
 
-			if (messageInfo.IsAnsweredAll)
+			if (CheckFlags (flags, Camel.CamelFlags.AnsweredAll))
 				indexable.AddProperty (Property.NewFlag ("fixme:isAnsweredAll"));
-#endif
-
+			
 			indexable.SetTextReader (PartHandler.GetReader (message));
 
 			return indexable;
@@ -483,10 +547,14 @@ namespace Beagle.Daemon.EvolutionMailDriver {
 	}
 
 	public class EvolutionMailIndexableGeneratorImap : EvolutionMailIndexableGenerator {
-		private static GConf.Client gconf_client = null;
+		private enum ImapBackendType {
+			Imap,
+			Imap4
+		};
 
 		private FileInfo summary_info;
 		private string imap_name;
+		private ImapBackendType backend_type;
 		private Camel.Summary summary;
 		private IEnumerator summary_enumerator;
 		private ICollection accounts;
@@ -521,8 +589,18 @@ namespace Beagle.Daemon.EvolutionMailDriver {
 		protected override bool Setup ()
 		{
 			string dir_name = summary_info.DirectoryName;
+			int imap_start_idx;
+
+			int idx = dir_name.IndexOf (".evolution/mail/imap4/");
+
+			if (idx >= 0) {
+				this.backend_type = ImapBackendType.Imap4;
+				imap_start_idx = idx + 22;
+			} else {
+				this.backend_type = ImapBackendType.Imap;
+				imap_start_idx = dir_name.IndexOf (".evolution/mail/imap/") + 21;
+			}
 
-			int imap_start_idx = dir_name.IndexOf (".evolution/mail/imap/") + 21;
 			string imap_start = dir_name.Substring (imap_start_idx);
 			this.imap_name = imap_start.Substring (0, imap_start.IndexOf ('/'));
 
@@ -560,11 +638,22 @@ namespace Beagle.Daemon.EvolutionMailDriver {
 				if (uid == null)
 					continue;
 
-				XmlNode imap_url = xmlDoc.SelectSingleNode ("//source/url");
+				XmlNode imap_url_node = xmlDoc.SelectSingleNode ("//source/url");
 
-				if (imap_url == null)
+				if (imap_url_node == null)
 					continue;
 
+				string imap_url = imap_url_node.InnerText;
+				// If there is a semicolon in the username part of the URL, it
+				// indicates that there's an auth scheme there.  We don't care
+				// about that, so remove it.
+				int user_end = imap_url.IndexOf ('@');
+				int semicolon = imap_url.IndexOf (';', 0, user_end + 1);
+
+				if (semicolon != -1)
+					imap_url = imap_url.Substring (0, semicolon) + imap_url.Substring (user_end);
+
+
 				// Escape out additional @s in the name.  I hate the class libs so much.
 				int lastIdx = this.imap_name.LastIndexOf ('@');
 				if (this.imap_name.IndexOf ('@') != lastIdx) {
@@ -572,7 +661,13 @@ namespace Beagle.Daemon.EvolutionMailDriver {
 					this.imap_name = toEscape.Replace ("@", "%40") + this.imap_name.Substring (lastIdx);
 				}
 
-				if (imap_url.InnerText.StartsWith ("imap://" + this.imap_name)) {
+				string backend_url_prefix;
+				if (this.backend_type == ImapBackendType.Imap)
+					backend_url_prefix = "imap";
+				else
+					backend_url_prefix = "imap4";
+
+				if (imap_url.StartsWith (backend_url_prefix + "://" + this.imap_name + "/")) {
 					this.account_name = uid;
 					break;
 				}
@@ -655,7 +750,10 @@ namespace Beagle.Daemon.EvolutionMailDriver {
 
 			if (this.summary == null) {
 				try {
-					this.summary = Camel.Summary.load (this.summary_info.FullName);
+					if (this.backend_type == ImapBackendType.Imap)
+						this.summary = Camel.Summary.LoadImapSummary (this.summary_info.FullName);
+					else
+						this.summary = Camel.Summary.LoadImap4Summary (this.summary_info.FullName);
 				} catch (Exception e) {
 					EvolutionMailQueryable.log.Warn ("Unable to index {0}: {1}", this.folder_name,
 									 e.Message);
@@ -716,6 +814,13 @@ namespace Beagle.Daemon.EvolutionMailDriver {
 			return reader;
 		}
 
+		// Kind of nasty, but we need the function.
+		[System.Runtime.InteropServices.DllImport("libglib-2.0.so.0")]
+		static extern int g_str_hash (string str);
+
+		// Stolen from deep within e-d-s's camel-data-cache.c  Very evil.
+		private const int CAMEL_DATA_CACHE_MASK = ((1 << 6) - 1);
+
 		public override Indexable GetNextIndexable ()
 		{
 			Indexable indexable = null;
@@ -724,44 +829,66 @@ namespace Beagle.Daemon.EvolutionMailDriver {
 
 			++this.count;
 
+			// Try to load the cached message data off disk
 			if (this.mapping[mi.uid] == null || (uint) mapping[mi.uid] != mi.flags) {
 				TextReader msgReader = null;
-				string msg_file = Path.Combine (summary_info.DirectoryName, mi.uid + ".");
+				string msg_file;
 
-				// FIXME - XXX
+				if (this.backend_type == ImapBackendType.Imap)
+					msg_file = Path.Combine (summary_info.DirectoryName, mi.uid + ".");
+				else {
+					// This is taken from e-d-s's camel-data-cache.c.  No doubt
+					// NotZed would scream bloody murder if he saw this here.
+					int hash = (g_str_hash (mi.uid) >> 5) & CAMEL_DATA_CACHE_MASK;
+					string cache_path = String.Format ("cache/{0:x}/{1}", hash, mi.uid);
+					msg_file = Path.Combine (summary_info.DirectoryName, cache_path);
+				}
+				
+				// FIXME - Filters really need to be rearchitected so that we can
+				// pass multiple attachments of different types into the indexable.
 				if (File.Exists (msg_file))
 					msgReader = GetMessageData (msg_file);
 
 				indexable = this.CamelMessageToIndexable (mi, msgReader);
 
 				this.mapping[mi.uid] = mi.flags;
-				++this.indexed_count;
+
+				if (indexable != null)
+					++this.indexed_count;
 			} 
 
-			this.deleted_list.Remove (mi.uid);
+			if (indexable != null) {
+				this.deleted_list.Remove (mi.uid);
+
+				// HACK: update your recipients
+				EvolutionMailQueryable.AddAsYourRecipient (indexable);
+			}
 
 			return indexable;
 		}
 
 		private Uri CamelMessageUri (Camel.MessageInfo message_info)
 		{
-			return EvolutionMailQueryable.EmailUri (this.account_name, this.folder_name, message_info.uid);			
+			return EvolutionMailQueryable.EmailUri (this.account_name, this.folder_name, message_info.uid);
 		}
 
 		private Indexable CamelMessageToIndexable (Camel.MessageInfo messageInfo, TextReader msgReader)
 		{
+			// Don't index messages flagged as junk
+			if (messageInfo.IsJunk)
+				return null;
+
 			Uri uri = CamelMessageUri (messageInfo);
 			Indexable indexable = new Indexable (uri);
 
 			indexable.Timestamp = messageInfo.Date;
 			indexable.Type = "MailMessage";
-			indexable.MimeType = "text/plain"; // FIXME
 
-			indexable.AddProperty (Property.NewKeyword ("dc:title", messageInfo.subject));
+			indexable.AddProperty (Property.New ("dc:title", messageInfo.subject));
+			indexable.AddProperty (Property.New ("fixme:subject",  messageInfo.subject));
 
 			indexable.AddProperty (Property.NewKeyword ("fixme:account",  this.imap_name));
                         indexable.AddProperty (Property.NewKeyword ("fixme:folder",   this.folder_name));
-			indexable.AddProperty (Property.NewKeyword ("fixme:subject",  messageInfo.subject));
                         indexable.AddProperty (Property.NewKeyword ("fixme:to",       messageInfo.to));
 			indexable.AddProperty (Property.NewKeyword ("fixme:from",     messageInfo.from));
                         indexable.AddProperty (Property.NewKeyword ("fixme:cc",       messageInfo.cc));
@@ -774,9 +901,15 @@ namespace Beagle.Daemon.EvolutionMailDriver {
 			if (messageInfo.sent != DateTime.MinValue)
 				indexable.AddProperty (Property.NewDate ("fixme:sentdate", messageInfo.sent));
 
-			if (this.folder_name == "Sent")
+			if (this.folder_name == "Sent") {
 				indexable.AddProperty (Property.NewFlag ("fixme:isSent"));
 
+				MapAddressStringToProperties (indexable, "fixme:sentTo", messageInfo.to);
+				MapAddressStringToProperties (indexable, "fixme:sentTo", messageInfo.cc);
+			} else {
+				MapAddressStringToProperties (indexable, "fixme:gotFrom", messageInfo.from);
+			}
+
 			if (messageInfo.IsAnswered)
 				indexable.AddProperty (Property.NewFlag ("fixme:isAnswered"));
 
@@ -800,6 +933,8 @@ namespace Beagle.Daemon.EvolutionMailDriver {
 
 			if (msgReader != null)
 				indexable.SetTextReader (msgReader);
+			else
+				indexable.NoContent = true;
 
 			return indexable;
 		}
