@@ -74,7 +74,10 @@ namespace Beagle.Daemon {
 		//    in the file system backend (it would be nice to have per-backend
 		//    versioning so that we didn't have to purge all indexes just
 		//    because one changed)
-		private const int VERSION = 8;
+		// 9: changed the way properties are stored, changed in conjunction
+		//    with sane handling of multiple properties on hits.
+		private const int MAJOR_VERSION = 9;
+		private int minor_version = 0;
 
 		private string top_dir;
 		private Hashtable pending_by_uri = UriFu.NewHashtable ();
@@ -84,9 +87,11 @@ namespace Beagle.Daemon {
 		private bool optimizing = false;
 		private int last_item_count = -1;
 
-		public LuceneDriver (string index_name)
+		public LuceneDriver (string index_name) : this (index_name, 0) { }
+
+		public LuceneDriver (string index_name, int index_version)
 		{
-			Setup (index_name);
+			Setup (index_name, index_version);
 		}
 
 		public string IndexDirectory {
@@ -113,9 +118,8 @@ namespace Beagle.Daemon {
 
 		/////////////////////////////////////////////////////
 
-
-		private void Setup (string index_name)
-		{
+		private void Setup (string index_name, int _minor_version)
+		{			
 			top_dir = Path.Combine (PathFinder.StorageDir, index_name); 
 			
 			string versionFile = Path.Combine (top_dir, "version");
@@ -128,15 +132,35 @@ namespace Beagle.Daemon {
 			bool fingerprintExists = File.Exists (fingerprintFile);
 			bool indexExists = File.Exists (indexTestFile);
 
+			if (_minor_version < 0)
+				_minor_version = 0;
+			minor_version = _minor_version;
+
 			// Check the index's version number.  If it is wrong,
 			// declare the index non-existent.
 			if (versionExists && indexExists) {
 				StreamReader sr = new StreamReader (versionFile);
 				string versionStr = sr.ReadLine ();
 				sr.Close ();
-				
-				if (versionStr != Convert.ToString (VERSION))
+
+				int old_major_version, old_minor_version;
+				int i = versionStr.IndexOf (".");
+
+				if (i != -1) {
+					old_major_version = Convert.ToInt32 (versionStr.Substring (0,i));
+					old_minor_version = Convert.ToInt32 (versionStr.Substring (i+1));
+				} else {
+					old_major_version = Convert.ToInt32 (versionStr);
+					old_minor_version = 0;
+				}
+
+				if (old_major_version != MAJOR_VERSION || old_minor_version != minor_version) {
+					log.Debug ("Version mismatch in {0}", index_name);
+					log.Debug ("Index has version {0}.{1}, expected {2}.{3}",
+						   old_major_version, old_minor_version,
+						   MAJOR_VERSION, minor_version);
 					indexExists = false;
+				}
 			}
 
 			// If there is no fingerprint file, declare the index
@@ -146,13 +170,14 @@ namespace Beagle.Daemon {
 
 			// If the index seems to exist but contains dangling locks,
 			// declare the index non-existent.
+
 			if (indexExists) {
 				DirectoryInfo lockDirInfo = new DirectoryInfo (lockDir);
 				if (! lockDirInfo.Exists)
 					indexExists = false;
 				else {
 					foreach (FileInfo info in lockDirInfo.GetFiles ()) {
-						if (info.Name.IndexOf (".lock") != -1) {
+						if (info.Name.IndexOf ("write.lock") != -1) {
 							indexExists = false;
 							break;
 						}
@@ -192,7 +217,7 @@ namespace Beagle.Daemon {
 
 				// Write out our version information
 				sw = new StreamWriter (versionFile, false);
-				sw.WriteLine ("{0}", VERSION);
+				sw.WriteLine ("{0}.{1}", MAJOR_VERSION, minor_version);
 				sw.Close ();
 			}
 
@@ -208,6 +233,11 @@ namespace Beagle.Daemon {
 				IndexWriter writer = new IndexWriter (Store, null, true);
 				writer.Close ();
 			}
+
+			if (Environment.GetEnvironmentVariable ("BEAGLE_OPTIMIZE_ON_STARTUP") != null) {
+				cycles_since_last_optimization = 1000; // this can't be zero, or nothing will happen
+				Optimize ();
+			}
 		}
 
 		/////////////////////////////////////////////////////
@@ -338,8 +368,8 @@ namespace Beagle.Daemon {
 			last_item_count = reader.NumDocs ();
 			reader.Close ();
 			watch.Stop ();
-			//Log.Debug ("Step #1: {0} {1} {2}", watch, pending_uris.Count,
-			//	   watch.ElapsedTime / pending_uris.Count);
+			Log.Debug ("Lucene Delete: {0} {1} {2}", watch, pending_uris.Count,
+				   pending_uris.Count / watch.ElapsedTime);
 
 
 			// Step #2: Write out the pending adds
@@ -348,6 +378,13 @@ namespace Beagle.Daemon {
 			foreach (Indexable indexable in pending_indexables) {
 				
 				Log.Debug ("+ {0}", indexable.DisplayUri);
+
+				try {
+					FilterFactory.FilterIndexable (indexable);
+				} catch (Exception e) {
+					Log.Error ("Unable to filter {0} (mimetype={1})", indexable.DisplayUri, indexable.MimeType);
+					Log.Error (e);
+				}
 				
 				Document doc = null;
 				try {
@@ -369,7 +406,8 @@ namespace Beagle.Daemon {
 			if (writer != null) 
 				writer.Close ();
 			watch.Stop ();
-			//Log.Debug ("Step #2: {0}", watch);
+			Log.Debug ("Lucene Add: {0} {1} {2}", watch, pending_indexables.Count,
+				   pending_indexables.Count / watch.ElapsedTime);
 			
 
 			// Step #3: Fire off an event telling what we just did.
@@ -426,7 +464,7 @@ namespace Beagle.Daemon {
 
 		// Returns the lowest matching score before the results are
 		// truncated.
-		public void DoQuery (QueryBody           body,
+		public void DoQuery (Query               query,
 				     IQueryResult        result,
 				     ICollection         search_subset, // should be internal uris
 				     ICollection         bonus_uris,    // should be internal uris
@@ -442,9 +480,9 @@ namespace Beagle.Daemon {
 			IndexReader reader = IndexReader.Open (Store);
 
 			LNS.Searcher searcher = new LNS.IndexSearcher (reader);
-			LNS.Query query = ToLuceneQuery (body, search_subset, bonus_uris);
+			LNS.Query lucene_query = ToLuceneQuery (query, search_subset, bonus_uris);
 
-			LNS.Hits hits = searcher.Search (query);
+			LNS.Hits hits = searcher.Search (lucene_query);
 			sw.Stop ();
 
 			t_lucene = sw.ElapsedTime;
@@ -576,10 +614,51 @@ namespace Beagle.Daemon {
 		// Code to map to/from Lucene data types
 		//
 
+		private void AddPropertyToLuceneDocument (Document doc, Property prop)
+		{
+			if (prop.Value == null)
+				return;
+
+			Field f;
+
+			if (prop.IsSearched) {
+				f = new Field (prop.IsKeyword ? "PropertyKeyword" : "PropertyText",
+					       prop.Value,
+					       false,              // never stored
+					       true,               // always index
+					       ! prop.IsKeyword);  // only tokenize non-keywords
+				doc.Add (f);
+			}
+
+			f = new Field (String.Format ("prop:{0}:{1}",
+						      prop.IsKeyword  ? "k" : "_",
+						      prop.Key),
+				       prop.Value,
+				       true,               // always store
+				       true,               // always index
+				       ! prop.IsKeyword);  // only tokenize non-keywords
+			doc.Add (f);
+		}
+
+		private Property FieldToProperty (Field field)
+		{
+			string name = field.Name ();
+			if (name.Length < 7 || ! name.StartsWith ("prop:"))
+				return null;
+
+			string key = name.Substring (7);
+			string value = field.StringValue ();
+			
+			if (name [5] == 'k')
+				return Property.NewKeyword (key, value);
+			else if (name [6] == 's')
+				return Property.NewUnsearched (key, value);
+			else
+				return Property.New (key, value);
+		}
+
 		private Document ToLuceneDocument (Indexable indexable)
 		{			
-			indexable.Build ();
-			
 			Document doc = new Document ();
 			Field f;
 			String str;
@@ -623,25 +702,8 @@ namespace Beagle.Daemon {
 				doc.Add (f);
 			}
 
-			f = Field.UnStored ("PropertiesText",
-					    indexable.TextPropertiesAsString);
-			doc.Add (f);
-
-			// FIXME: We shouldn't apply stemming, etc. when dealing
-			// with this field.
-			f = Field.UnStored ("PropertiesKeyword",
-					    indexable.KeywordPropertiesAsString);
-			doc.Add (f);
-			
-			// FIXME: We need to deal with duplicate properties in some
-			// sort of sane way.
-			foreach (Property prop in indexable.Properties) {
-				if (prop.Value != null) {
-					f = Field.Keyword (ToLucenePropertyKey (prop.Key),
-							   prop.Value);
-					doc.Add (f);
-				}
-			}
+			foreach (Property prop in indexable.Properties)
+				AddPropertyToLuceneDocument (doc, prop);
 			
 			return doc;
 		}
@@ -659,6 +721,7 @@ namespace Beagle.Daemon {
 				Uri uri = original_uri;
 				if (remapper != null)
 					uri = remapper (uri);
+				//Logger.Log.Debug ("ToUriQuery: {0} => {1}", original_uri, uri);
 				Term term = new Term ("Uri", uri.ToString ()); // FIXME: Do we need some UriFu here?
 				LNS.Query term_query = new LNS.TermQuery (term);
 				query.Add (term_query, false, false);
@@ -679,70 +742,50 @@ namespace Beagle.Daemon {
 		{
 			return ToUriQuery (list_of_uris, null);
 		}
-		
-		private LNS.Query ToCoreLuceneQuery (QueryBody body, string field)
+
+		static private LNS.Query NewTokenizedQuery (string field, string text)
 		{
-			LNS.BooleanQuery luceneQuery = null;
-			foreach (string text_orig in body.Text) {
-				string text = text_orig;
+			ArrayList tokens = new ArrayList ();
 
-				if (text == null || text == "")
-					continue;
-
-				bool minus_sign = false;
-				if (text [0] == '-') {
-					text = text.Substring (1);
-					minus_sign = true;
-				}
-
-				// Use the analyzer to extract the query's tokens.
-				// This code is taken from Lucene's query parser.
-				// We use the standard Analyzer.
-				TokenStream source = LuceneDriver.Analyzer.TokenStream (field, new StringReader (text));
-				ArrayList tokens = new ArrayList ();
-
-				while (true) {
-					Lucene.Net.Analysis.Token t;
-					try {
-						t = source.Next ();
-					} catch (IOException) {
-						t = null;
-					}
-					if (t == null)
-						break;
-					tokens.Add (t.TermText ());
-				}
+			// Use the analyzer to extract the query's tokens.
+			// This code is taken from Lucene's query parser.
+			// We use the standard Analyzer.
+			TokenStream source = LuceneDriver.Analyzer.TokenStream (field, new StringReader (text));
+			while (true) {
+				Lucene.Net.Analysis.Token t;
 				try {
-					source.Close ();
-				} catch (IOException) { 
-					// ignore
+					t = source.Next ();
+				} catch (IOException) {
+					t = null;
 				}
-
-				LNS.Query q = null;
-				if (tokens.Count == 1) {
-					Term t = new Term (field, (string) tokens [0]);
-					q = new LNS.TermQuery (t);
-				} else if (tokens.Count > 1) {
-					q = new LNS.PhraseQuery ();
-					foreach (string tokenStr in tokens) {
-						Term t = new Term (field, tokenStr);
-						((LNS.PhraseQuery) q).Add (t);
-					}
-				}
-
-				if (q != null) {
-					if (luceneQuery == null)
-						luceneQuery = new LNS.BooleanQuery ();
-					luceneQuery.Add (q, !minus_sign, minus_sign);
+				if (t == null)
+					break;
+				tokens.Add (t.TermText ());
+			}
+			try {
+				source.Close ();
+			} catch (IOException) { 
+				// ignore
+			}
+			
+			LNS.Query q = null;
+			if (tokens.Count == 1) {
+				Term t = new Term (field, (string) tokens [0]);
+				q = new LNS.TermQuery (t);
+			} else if (tokens.Count > 1) {
+				q = new LNS.PhraseQuery ();
+				foreach (string tokenStr in tokens) {
+					Term t = new Term (field, tokenStr);
+					((LNS.PhraseQuery) q).Add (t);
 				}
 			}
-			return luceneQuery;
 
+			return q;
 		}
 
 		// search_subset limits the score of our search to that set of Uris
 		// bonus_uris are always matched by the query
-		private LNS.Query ToLuceneQuery (QueryBody body,
+		private LNS.Query ToLuceneQuery (Query query,
 						 ICollection search_subset,
 						 ICollection bonus_uris)
 		{
@@ -751,53 +794,92 @@ namespace Beagle.Daemon {
 			LNS.Query        bonus_uris_query = null;
 			LNS.BooleanQuery mime_type_query = null;
 			LNS.BooleanQuery hit_type_query = null;
+
+			body_query = new LNS.BooleanQuery ();
+
+			bool used_any_part = false;
 			
-			if (body.Text.Count > 0) {
-
-				body_query = new LNS.BooleanQuery ();
-
-				LNS.Query q;
+			foreach (QueryPart part in query.Parts) {
 				
-				q = ToCoreLuceneQuery (body, "PropertiesText");
-				if (q != null) {
-					q.SetBoost (2.5f);
-					body_query.Add (q, false, false);
+				LNS.BooleanQuery part_query = new LNS.BooleanQuery ();
+				LNS.Query part_query_override = null;
+				LNS.Query subquery = null;
+
+				bool used_this_part = false;
+				
+				if (part.TargetIsAll || part.TargetIsText) {
+					
+					subquery = NewTokenizedQuery ("Text", part.Text); 
+					if (subquery != null) {
+						part_query.Add (subquery, false, false);
+						used_this_part = true;
+					}
+
+					subquery = NewTokenizedQuery ("HotText", part.Text);
+					if (subquery != null) {
+						subquery.SetBoost (1.75f);
+						part_query.Add (subquery, false, false);
+						used_this_part = true;
+					}
 				}
 
-				q = ToCoreLuceneQuery (body, "PropertiesKeyword");
-				if (q != null) {
-					q.SetBoost (2.5f);
-					body_query.Add (q, false, false);
+				if (part.TargetIsAll || part.TargetIsProperties) {
+					subquery = NewTokenizedQuery ("PropertyText", part.Text);
+					if (subquery != null) {
+						subquery.SetBoost (1.75f);
+						part_query.Add (subquery, false, false);
+						used_this_part = true;
+					}
+				}
+
+				if (part.TargetIsSpecificProperty) {
+					
+					string prop_name;
+					prop_name = String.Format ("prop:{0}:{1}",
+								   part.IsKeyword ? 'k' : '_',
+								   part.Target);
+					
+					if (part.IsKeyword) {
+						Term term = new Term (prop_name, part.Text);
+						subquery = new LNS.TermQuery (term);
+					} else {
+						subquery = NewTokenizedQuery (prop_name, part.Text);
+					}
+					
+					// Instead of the boolean query, just use the subquery.
+					if (subquery != null) {
+						part_query_override = subquery;
+						used_this_part = true;
+					}
 				}
 				
-				q = ToCoreLuceneQuery (body, "HotText");
-				if (q != null) {
-					q.SetBoost (1.75f);
-					body_query.Add (q, false, false);		
-				}
-				
-				q = ToCoreLuceneQuery (body, "Text");
-				if (q != null) {
-					body_query.Add (q, false, false);
+				if (used_this_part) {
+					if (part_query_override == null)
+						part_query_override = part_query;
+					body_query.Add (part_query_override, part.IsRequired, part.IsProhibited);
+					used_any_part = true;
 				}
 			}
 
+			if (! used_any_part)
+				return null;
+		
 			search_subset_query = ToUriQuery (search_subset, null);
 
 			bonus_uris_query = ToUriQuery (bonus_uris, null);
 				
-			if (body.MimeTypes.Count > 0) {
+			if (query.MimeTypes.Count > 0) {
 				mime_type_query = new LNS.BooleanQuery ();
-				foreach (string mime_type in body.MimeTypes) {
+				foreach (string mime_type in query.MimeTypes) {
 					Term t = new Term ("MimeType", mime_type);
 					LNS.Query q = new LNS.TermQuery (t);
 					mime_type_query.Add (q, false, false);
 				}
 			}
 
-			if (body.HasHitTypes) {
+			if (query.HasHitTypes) {
 				hit_type_query = new LNS.BooleanQuery ();
-				foreach (string hit_type in body.HitTypes) {
+				foreach (string hit_type in query.HitTypes) {
 					Term t = new Term ("Type", hit_type);
 					LNS.Query q = new LNS.TermQuery (t);
 					hit_type_query.Add (q, false, false);
@@ -892,9 +974,9 @@ namespace Beagle.Daemon {
 			hit.ScoreRaw = score;
 			
 			foreach (Field ff in doc.Fields ()) {
-				string key = FromLucenePropertyKey (ff.Name ());
-				if (key != null)
-					hit [key] = ff.StringValue ();
+				Property prop = FieldToProperty (ff);
+				if (prop != null)
+					hit.AddProperty (prop);
 			}
 			
 			return hit;
@@ -1067,20 +1149,6 @@ namespace Beagle.Daemon {
 			return Convert.ToInt64 (str);
 		}
 
-		const string propPrefix = "prop:";
-
-		private string ToLucenePropertyKey (string key)
-		{
-			return propPrefix + key;
-		}
-
-		private string FromLucenePropertyKey (string key)
-		{
-			if (key.StartsWith (propPrefix))
-				return key.Substring (propPrefix.Length);
-			return null;
-		}
-
 		/////////////////////////////////////////////////////
 
 		// Expose some information for debugging and analytical purposes.
